#!/usr/bin/env python3
# coding=utf-8
from array import array
from typing import List, Dict

from .simulator import Sim
from .events import BaseEvent


class RateEstimator(object):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.all_estimators: Dict[int, List['RateEstimator']] = dict()
    
    def __init__(self, sim: Sim, period: int = 500000):
        self.period = int(period)
        self.sim = sim
        estimator_props = sim.props.rate_estimator
        assert self.period > 0
        
        if self.period not in estimator_props.all_estimators:
            estimator_props.all_estimators[self.period] = []
            sim.schedule_event(RateEstimatorUpdateEvent(self))
        estimator_props.all_estimators[self.period].append(self)
    
    def packet_arrival(self):
        pass
    
    def get_estimated_rate(self):
        pass
    
    def update_rate(self):
        pass


class RateEstimatorUpdateEvent(BaseEvent):
    
    def __init__(self, estimator: RateEstimator):
        super().__init__(estimator.sim.currentTime + estimator.period)
        self.estimator = estimator
        self.ignoreWhenFinished = True
    
    def process_event(self):
        for rate_estimator in self.estimator.sim.props.rate_estimator.all_estimators[self.estimator.period]:
            rate_estimator.update_rate()
        self.estimator.sim.schedule_event(RateEstimatorUpdateEvent(estimator=self.estimator))


class EWMA(RateEstimator):
    
    def __init__(self, sim: Sim, alpha: float = 0.06, period: int = 5000, buckets: int = 20):
        super().__init__(sim, period=period)
        self.expected_size = buckets
        self.value = 0
        self.alpha = alpha
        assert 1 > self.alpha > 0
        
        assert 0 < self.expected_size < 100000
        self.buckets = array('I', [0])
        self.pos_current_bucket = 0
    
    def packet_arrival(self):
        self.value += 1
    
    def get_estimated_rate(self):
        v = float(self.buckets[self.pos_current_bucket])
        length = len(self.buckets)
        next_pos = (self.pos_current_bucket + 1) % length
        while next_pos != self.pos_current_bucket:
            v = v * (1 - self.alpha) + self.alpha * float(self.buckets[next_pos])
            next_pos = (next_pos + 1) % length
        return v / (self.period / 1000000)
    
    def update_rate(self):
        if len(self.buckets) < self.expected_size:
            self.buckets.append(self.value)
            self.pos_current_bucket = len(self.buckets) - 1
        else:
            self.pos_current_bucket = (self.pos_current_bucket + 1) % len(self.buckets)
            self.buckets[self.pos_current_bucket] = self.value
        self.value = 0


class DRE(RateEstimator):
    """Discounting Rate Estimator (DRE)
    should result in: X is proportional to the rate of traffic
    more precisely, if the traffic rate is R, then X ≈ R · τ, where τ = Tdre/α
    https://people.csail.mit.edu/alizadeh/papers/conga-techreport.pdf"""
    
    def __init__(self, sim: Sim, alpha: float = 0.125, period: int = 500000):
        super().__init__(sim, period)
        self.value = 0
        self.alpha = alpha
        assert 1 > self.alpha > 0
        
        self.tau = (1000000 / self.period) / self.alpha
    
    def packet_arrival(self):
        self.value += 1
    
    def update_rate(self):
        self.value *= (1 - self.alpha)
    
    def get_dre(self):
        return self.value
    
    def get_estimated_rate(self):
        return self.value / self.tau
    
    def get_congestion_metric(self, capacity):
        return self.value / (self.tau * capacity)
    
    def get_tau(self):
        return self.tau



#!/usr/bin/env python3
# coding=utf-8
from .model.core import Flow
from .model.sff import SFF
from .model.sfi import SFI
from .scheduler.core import BaseScheduler
from .simulator import Sim


class BG_Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def colour_print(value, warning: bool):
    return f'{BG_Colors.WARNING if warning else BG_Colors.OKGREEN}{value}{BG_Colors.ENDC}'


def print_sim_snapshot(sim: Sim, per_sfi_stats: bool = False):
    if sim.stats is None:
        sim.calculate_simple_statistics()
    
    queued_packets_sff = sum([sff.get_number_of_queued_packets() for sff in sim.props.sff.allSFFs.values()])
    print(f' packets queued among all SFFs:'
          f' {colour_print(queued_packets_sff, queued_packets_sff > 0)}')
    print(
        f' SFI status: '
        f'{len([sfi for sfi in sim.props.sfi.all_sfi.values() if sfi.free])}'
        f'/{len(sim.props.sfi.all_sfi)} free/all')
    queued_packets_sfi = [len(sfi.queue) for sfi in sim.props.sfi.all_sfi.values()]
    if per_sfi_stats:
        print(f' packets queued among all SFIs:')
        for id, of_type, packets in [(sfi.id, sfi.of_type, len(sfi.queue)) for sfi in sim.props.sfi.all_sfi.values()]:
            print(f'\tid {id}: {packets} of type {of_type}')
    else:
        print(f' packets queued among all SFIs:'
              f'{colour_print(sum(queued_packets_sfi), sum(queued_packets_sfi) > 0)}')
    
    print(f' idle time of all servers: {sim.stats["server_idle_time"]}, '
          f'on avg {sim.stats["server_idle_time_ratio"]}% each server over time')
    
    print(f' statistics\n\t{sim.props.packet.statsPacketsRejectedProcessingDelay}: timeout')
    print(f'\t{sim.props.packet.statsPacketsRejectedSchedule}: scheduler reject')
    print(f'\t{sim.props.packet.statsPacketsSuccessfulProcessed}: successful processed')
    print(f'\t{sim.props.packet.statsPacketsTotalCount}: total')


def validate_path(path: list, start_sff: SFF):
    last_sff: SFF = start_sff
    last_hop = start_sff
    sim = start_sff.sim
    
    for entry_type, entry in path:
        if entry_type == SFF.__name__:
            assert entry.sim == sim
            # last hop was SFI?
            if last_hop.__class__.__name__ == SFI.__name__:
                # check if SFI belongs to this SFF
                assert (last_hop.sffId == entry.id)
            else:
                # last hop is SFF, so check if there exists a connection
                assert (SFF.check_if_connection_exists(sim, last_hop.id, entry.id))
            last_sff = entry
        else:
            # entry if sfi,
            # check if its part of the last SFF
            assert (entry.sffId == last_sff.id)
        
        last_hop = entry


def print_simple_stats(sim: Sim, per_sfi_stats: bool = False):
    sim.calculate_simple_statistics()
    print("simple stats")
    print(f"\tsuccess rate: \t\t{sim.stats['success_rate']}")
    print(f"\tservice_quality: \t{sim.stats['service_quality']}")
    
    print(
        f"\t packets fed to the simulator:\t\t\t{sim.props.packet.statsPacketsTotalCount} "
        f"({sim.props.flow.lastId} flows, {sim.props.flow.sfc_next_free_class} traffic classes)")
    print("\t packets rejected during scheduling:\t\t{0}".format(
        str(sim.props.packet.statsPacketsRejectedSchedule)))
    print("\t packets dropped because of delay max out:\t{0}".format(
        str(sim.props.packet.statsPacketsRejectedProcessingDelay)))
    print(
        f"\t total scheduling attempts: \t\t\t{sum([s.scheduler.scheduling_attempts for s in sim.props.sff.allSFFs.values()])}")
    print(f"\t packets successful processed:\t\t\t{sim.props.packet.statsPacketsSuccessfulProcessed} " +
          f"({sim.props.packet.counter_packet_after_workload_end_in_system_no_timeout} packets still in "
          f"the simulation when workload was over, but delay < QoS deadline)")
    if (
            sim.props.packet.statsPacketsSuccessfulProcessed - sim.props.packet.counter_packet_after_workload_end_in_system_no_timeout) > 0:
        print("\t successful, avg delay/max delay:\t\t{0}".format(
            str(sim.props.packet.statsRatiosQos / (
                    sim.props.packet.statsPacketsSuccessfulProcessed - sim.props.packet.counter_packet_after_workload_end_in_system_no_timeout))))
    print("\t total delay among all packets:\t\t\t{0}".format(
        str(sim.props.packet.statsSumDelay)))
    sim.print_sim_snapshot(per_sfi_stats=per_sfi_stats)


def print_infrastructure(sim: Sim, with_snap: bool = False, per_sfi_stats: bool = False):
    print("SFFs:")
    sff: SFF
    for sff in sim.props.sff.allSFFs.values():
        print(f" SFF {sff.id}")
        if sim.props.sff.linkBwCap is not None:
            print(
                f"\tconnected to: {[other.id for other in sim.props.sff.allSFFs.values() if SFF.check_if_connection_exists(sim, sff.id, other.id)]}")
        print(f"\t{len(sff.SFIsPerType)} sf types: {sorted(sff.SFIsPerType.keys())}")
        print(f'\t{sum([len(x) for x in sff.SFIsPerType.values()])} SFIs: '
              f'{["sf " + str(sf) + " -> " + str(len(sff.SFIsPerType[sf])) + " sfis" for sf in sff.SFIsPerType]}')
        print(f"\t{len(sff.servers)} servers: {sorted([s.id for s in sff.servers])}")
        print(f"\t{sff.scheduler}")
        if with_snap:
            sched: BaseScheduler = sff.scheduler
            found = {}
            if sched.requires_queues_per_class():
                for packet_class in sff.packet_queue_per_class:
                    if 0 < len(sff.packet_queue_per_class[packet_class]):
                        found_sfc, pos = Flow.debug_get_sfc_identifier_and_pos_of_packet_class(sim, packet_class)
                        
                        for rel in range(-5, 5, 1):
                            print(
                                f'.. {packet_class + rel} -> {Flow.debug_get_sfc_identifier_and_pos_of_packet_class(sim, packet_class + rel)}')
                        
                        sf, eoc = Flow.get_sf_and_eoc_of(sim, found_sfc, pos)
                        found_sfc = f'{found_sfc}/{pos}/{sf}'
                        if found_sfc not in found:
                            found[found_sfc] = len(sff.packet_queue_per_class[packet_class])
                        else:
                            found[found_sfc] += len(sff.packet_queue_per_class[packet_class])
            for sfc in found:
                print(f'\t  {found[sfc]} packets of sfc: {sfc}')
            print_sim_snapshot(sim, per_sfi_stats)
    
    print("Servers:")
    for server in sim.props.server.all_servers:
        print(f" Server {server.id} hosting {len(server.SFIs)} SFIs of {len(server.SFF_ids)} SFFs")
        print(f"\t {server.SFIs}")
    
    print("SFI statistics")
    
    # for each SF type, how many instances do we have?
    # for each SF type, what is the processing capacity (total/effective shared)
    sfi_props: SFI.Props = sim.props.sfi
    # sfi count, raw capacity, shared capacity
    sfi_capacity = {sf: [0, 0, 0] for sf in range(len(sfi_props.processingRateOfSfType))}
    for sfi in sfi_props.all_sfi.values():
        sfi_capacity[sfi.of_type][0] += 1
        sfi_capacity[sfi.of_type][1] += sfi.server.processing_cap
        sfi_capacity[sfi.of_type][2] += sfi.server.processing_cap / len(sfi.server.SFIs)
    
    for sf in sfi_capacity:
        print(f"\t{sf}: {sfi_capacity[sf][0]} SFIs, {sfi_capacity[sf][1]}/{round(sfi_capacity[sf][2], 1)} capacity")
    
    print("SFI / SFF statistics")
    sff_count = len(sim.props.sff.allSFFs)
    for sf in sfi_capacity:
        print(f"\t{sf}: {round(sfi_capacity[sf][0] / sff_count, 1)} SFIs, "
              f"{round(sfi_capacity[sf][1] / sff_count, 1)}/{round(sfi_capacity[sf][2] / sff_count, 1)} capacity")
    
    return sfi_capacity



#!/usr/bin/env python3
# coding=utf-8

import time
from itertools import accumulate
from typing import Dict, List

from ..events import BaseEvent
from ..simulator import Sim, SchedulingFailure

from ..model.server import Server, ServerCpuPolicy
from ..model.sff import SFF
from ..model.sfi import SFI
from ..model.core import Flow, Packet

from ..rate_estimator import RateEstimator, EWMA


class ACP(object):
    
    def __init__(self, scheduler: 'BaseScheduler'):
        self.scheduler = scheduler
        self.dist_information = None  # sff_id -> ∀ sf unrolled the ratio of load / service
        #                                         time of information
        
        self.detour = 0
        self.non_detour = 0
    
    def check_packet_to_process_not_locally_and_update_path(self, packet: Packet):
        if self.scheduler.static_sff_rates_per_sf_sorted_sff is None:
            self.scheduler.update_cum_weights_of_sff_rates()
        
        forward_to_neighbor = False
        
        # forward because we cannot serve?
        next_sf = packet.toBeVisited[0]
        if next_sf not in self.scheduler.mySFF.SFIsPerType or len(self.scheduler.mySFF.SFIsPerType[next_sf]) == 0:
            forward_to_neighbor = True
        
        # if we are not forced to forward to a neighbor, check the load
        if not forward_to_neighbor and len(self.scheduler.static_sff_rates_per_sf_sorted_sff[next_sf]) > 0:
            # get the load of the affected queue
            arrival_rate = self.scheduler.get_arrival_rate_estimate(next_sf)
            service_rate = self.scheduler.mySFF.service_rate_per_sf[next_sf]
            assert service_rate > 0
            
            threshold_low = self.scheduler.admission_control_threshold_low
            threshold_high = self.scheduler.admission_control_threshold_high
            load = arrival_rate / service_rate
            
            # if self.scheduler.mySFF.id == 0:
            #     print(arrival_rate, load)
            
            if threshold_high > load > threshold_low:
                # load is still ok, but we forward with a certain ratio
                to_test = (load - threshold_low) / (threshold_high - threshold_low)
                if self.scheduler.sim.random.random() <= to_test:
                    forward_to_neighbor = True
            
            elif load >= threshold_high:
                # load is above higher threshold, forward packet
                forward_to_neighbor = True
        
        if forward_to_neighbor:
            # check if we are the only guys who have a sfi of certain type, if so, ignore local load and process it locally
            if (next_sf not in self.scheduler.static_sff_rates_per_sf_sorted_sff or
                    next_sf not in self.scheduler.static_sff_rates_per_sf_cum_weights or
                    len(self.scheduler.static_sff_rates_per_sf_sorted_sff[next_sf]) == 0):
                # there is no one with a SFI, so do not search for a path
                pass
            elif (next_sf in self.scheduler.static_sff_rates_per_sf_sorted_sff and
                  next_sf in self.scheduler.static_sff_rates_per_sf_cum_weights and
                  len(self.scheduler.static_sff_rates_per_sf_sorted_sff[next_sf]) == 1 and
                  self.scheduler.static_sff_rates_per_sf_sorted_sff[next_sf] == self.scheduler.mySFF.id):
                # we are the only ones with such a SFI, so process locally (ignore load)
                forward_to_neighbor = False
            else:
                self.recommend_forwarding_neighbor_and_update_path(packet=packet)
        
        return forward_to_neighbor
    
    def recommend_forwarding_neighbor_and_update_path(self, packet: Packet):
        next_sf = packet.toBeVisited[0]
        # select a neighbor
        assert next_sf in self.scheduler.static_sff_rates_per_sf_cum_weights
        assert 0 < len(self.scheduler.static_sff_rates_per_sf_cum_weights[next_sf])
        target_sff_id = self.scheduler.sim.random.choices(
            self.scheduler.static_sff_rates_per_sf_sorted_sff[next_sf],
            cum_weights=self.scheduler.static_sff_rates_per_sf_cum_weights[next_sf],
            k=1)[0]
        
        path_to_other_sff = SFF.get_multi_hop_path_for(self.scheduler.sim, self.scheduler.mySFF.id, target_sff_id)
        for next_sff in path_to_other_sff:
            packet.fullPath.append((SFF.__name__, self.scheduler.sim.props.sff.allSFFs[next_sff]))


class BaseScheduler(object):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.drop_packet_after_scheduling_attempts = 50
    
    def __repr__(self):
        return f'{self.__class__.__name__}{vars(self)}'
    
    def __init__(self, sim: Sim,
                 incremental: bool = True,
                 oracle=True,
                 activate_acp: bool = False,
                 admission_control_threshold_low: float = 0.8,
                 admission_control_threshold_high: float = 1.1):
        self.incremental = incremental
        self.oracle = oracle
        self.mySFF: SFF = None
        self.sim: Sim = sim
        self.time_scheduling_starts = None
        self.acp = ACP(self) if activate_acp else None
        
        self.admission_control_threshold_low = admission_control_threshold_low
        self.admission_control_threshold_high = admission_control_threshold_high
        
        assert admission_control_threshold_low < admission_control_threshold_high
        
        # we have two main data structures with information about all the SFIs available at other SFFs
        # and at all SFFs (other+my sff)
        # __sff__rates holds for each sf type, the aggregates sfi rate of all SFFs
        self.static_sff_rates_per_sf_cum_weights: Dict[int, List[float]] = None
        self.static_sff_rates_per_sf_sorted_sff: Dict[int, List[int]] = None
        
        # __sfi__rates holds for each sf type, the sfi rates
        self.static_sfi_rates_per_sf_cum_weights: Dict[int, List[float]] = None
        self.static_sfi_rates_per_sf_sorted_sfi: Dict[int, List[int]] = None
        
        self.scheduling_attempts = 0
        
        self.rate_estimator: Dict[int, RateEstimator] = {}
    
    def inform_rate_estimator_about_packet_arrival(self, packet: Packet):
        of_sf_type = packet.toBeVisited[0]
        if of_sf_type not in self.rate_estimator:
            self.rate_estimator[of_sf_type] = EWMA(self.sim)
        
        self.rate_estimator[of_sf_type].packet_arrival()
    
    def update_cum_weights_of_sff_rates(self, include_own_sff: bool = False):
        self.static_sff_rates_per_sf_cum_weights = {}
        self.static_sff_rates_per_sf_sorted_sff = {}
        # get all sff instanes, but not my sff, and sort them based on the sff_id
        # we have to sort them, so that the simulator runs deterministic (dictionaries are not sorted deterministic)
        list_of_sff = [self.sim.props.sff.allSFFs[sff_id] for sff_id in sorted(self.sim.props.sff.allSFFs.keys()) if
                       (sff_id != self.mySFF.id or include_own_sff)]
        
        # for each sf type
        for sf in range(len(self.sim.props.sfi.processingRateOfSfType)):
            # get the rate of SFF for the given sf, i.e., we take for each SFF the static! rate of its SFIs.
            # this is the upper limit, the SFF is capable of serving for this SF.
            # it is the upper limit, because we take the rate we would get, if all SFIs (of the given sf type)
            # of this SFF are working with full server power.
            rates = [sff.service_rate_per_sf[sf] for sff in list_of_sff
                     if sf in sff.service_rate_per_sf]
            # we take the cum weights of these rates and store them for later
            self.static_sff_rates_per_sf_sorted_sff[sf] = [sff.id for sff in list_of_sff
                                                           if sf in sff.service_rate_per_sf]
            self.static_sff_rates_per_sf_cum_weights[sf] = list(accumulate(rates))
            assert len(self.static_sff_rates_per_sf_sorted_sff[sf]) == len(self.static_sff_rates_per_sf_cum_weights[sf])
            
            # if len(self.static_sff_rates_per_sf_sorted_sff[sf]) == 0:
            #     raise NameError(f"there is no other SFF which hosts a SFI of type {sf}")
        # check that we have for each sf type some entries
        assert len(self.static_sff_rates_per_sf_cum_weights) == len(self.sim.props.sfi.processingRateOfSfType)
    
    def update_cum_weights_of_sfi_rates(self, include_own_sff: bool = False):
        # analogously to update_cum_weights_of_sff_rates, this method creates the data structure with respect ot SFIs
        sfi_props: SFI.Props = self.sim.props.sfi
        self.static_sfi_rates_per_sf_cum_weights = {}
        # get all sfis
        list_of_sfi = [sfi_props.all_sfi[sfi_id] for sfi_id in sorted(sfi_props.all_sfi.keys()) if
                       (sfi_props.all_sfi[sfi_id].sffId != self.mySFF.id or include_own_sff)]
        self.static_sfi_rates_per_sf_sorted_sfi = {}
        for sf in range(len(sfi_props.processingRateOfSfType)):
            rates = [sfi.get_expected_processing_rate() for sfi in list_of_sfi if sfi.of_type == sf]
            self.static_sfi_rates_per_sf_cum_weights[sf] = list(accumulate(rates))
            self.static_sfi_rates_per_sf_sorted_sfi[sf] = [sfi.id for sfi in list_of_sfi if sfi.of_type == sf]
            assert len(self.static_sfi_rates_per_sf_sorted_sfi[sf]) == len(self.static_sfi_rates_per_sf_cum_weights[sf])
        # check that we have for each sf type some entries
        assert len(self.static_sfi_rates_per_sf_cum_weights) == len(self.sim.props.sfi.processingRateOfSfType)
    
    def assign_sff(self, sff: SFF):
        self.mySFF = sff
        assert self.sim == sff.sim
    
    def requires_queues_per_class(self):
        # whether the scheduler asks for a queue per class at the sff, or a single queue
        return False
    
    def get_arrival_rate_estimate(self, of_sf_type: int):
        if of_sf_type in self.rate_estimator:
            return self.rate_estimator[of_sf_type].get_estimated_rate()
        return 0
    
    def handle_packet_arrival(self, packet: Packet):
        # notification for packet arrival
        # is there something to do for us?
        if len(packet.fullPath) > packet.pathPosition:
            print(Packet.debug_packet(packet))
            raise NameError(f'something is going wrong. a packet was given to the scheduler, '
                            f'but the packet has some steps in its path!')
        
        # first we need to check with the ACP if we should enqueue the packet or if we should process it locally
        
        if self.acp is not None:
            if self.acp.check_packet_to_process_not_locally_and_update_path(packet):
                # forward packet
                
                # get the packet from the scheduler's queue
                if self.requires_queues_per_class():
                    popped_packet = self.mySFF.packet_queue_per_class[Flow.get_packet_class_of_packet(packet)].pop()
                else:
                    popped_packet = self.mySFF.packet_queue.pop()
                
                assert popped_packet == packet
                
                # fake queueing time, so reset the packet's timer
                packet.timeQueueScheduling += packet.get_delta_of_time_mark()
                
                # if we forward, check if there are some other guys who are able to process the packet
                next_sf = packet.toBeVisited[0]
                if (next_sf not in self.static_sff_rates_per_sf_sorted_sff or
                        len(self.static_sff_rates_per_sf_sorted_sff[next_sf]) == 0):
                    # there is no other sff, so simply reject the packet
                    packet.reject()
                    raise SchedulingFailure(f'no sfi found for sf type {next_sf}')
                
                if packet.seenByScheduler > self.sim.props.base_scheduler.drop_packet_after_scheduling_attempts:
                    packet.drop_timed_out()
                    return
                
                # inform my sff to forward the packet
                self.mySFF.handle_packet_from_scheduler(packet)
                return
        
        self.inform_rate_estimator_about_packet_arrival(packet)
        
        # handle packet by the schedulers logic
        self.apply_scheduling_logic_for_packet(packet)
    
    def apply_scheduling_logic_for_packet(self, packet: Packet):
        raise NotImplemented
    
    def trigger_scheduling_logic(self) -> bool:
        return False
    
    def notify_sfi_finished_processing_of_packet(self, sfi: SFI, packet: Packet):
        # notification that a sfi finished processing
        pass
    
    def applies_round_robin(self):
        # this method is used for test cases, to run RR tests if the scheduler applies RR logic
        return False
    
    def is_always_able_to_build_full_path(self):
        # this method is used for test cases, to check whether this scheduler should be alble to build always
        # a full path, if requested (incremental=False)
        return True
    
    def supports_cpu_policy(self, cpu_policy: ServerCpuPolicy):
        # this method is used for test cases, to check whether this scheduler should be able to handle the given CPU policy
        return True
    
    def mark_time_scheduling_starts(self):
        assert self.time_scheduling_starts is None
        self.time_scheduling_starts = time.time()
    
    def reset_timer(self):
        assert self.time_scheduling_starts is not None
        self.time_scheduling_starts = None
    
    def get_time_delta_of_scheduling(self):
        delta = time.time() - self.time_scheduling_starts
        self.time_scheduling_starts = None
        return delta
    
    def get_load_of_sfis_of_sf(self, sf: int):
        raise NameError("not implemented")


class RejectScheduler(BaseScheduler):
    
    def apply_scheduling_logic_for_packet(self, packet: Packet):
        self.mySFF.packet_queue.pop()
        packet.reject()


class DoSchedulingEvent(BaseEvent):
    def __init__(self, delay, scheduler: 'BaseScheduler'):
        super().__init__(scheduler.sim.currentTime + delay)
        self.scheduler = scheduler
    
    def process_event(self):
        if self.scheduler.sim.DEBUG:
            print("doScheduling event triggers scheduler {0} of sff {1}".
                  format(self.scheduler, self.scheduler.mySFF.id))
        self.scheduler.trigger_scheduling_logic()



#!/usr/bin/env python3
# coding=utf-8


__all__ = ["model",
           "scheduler",
           "measurement",
           "workload",
           "sanity",
           "simulator",
           "events"]

import json
import os

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "meta.json"), "r") as f:
    meta = json.load(f)
    __version__ = meta["version"]
    __author__ = meta["author"]
    __license__ = meta["license"]
    __title__ = meta["title"]
    __copyright__ = meta["copyright"]

from .simulator import *
from .measurement import *
from .workload import *
from .sanity import *
from .events import *
from .scheduler import *
from .model import *



#!/usr/bin/env python3
# coding=utf-8
import random
from collections import deque
from typing import Dict, Generator, Deque, List

import numpy as np

from .model.core import Flow, Packet
from .model.sfi import SFI
from .simulator import Sim


class WorkloadGenerator(object):
    
    def __init__(self, sim: Sim, workload_rand: random.Random, config: Dict):
        self.sim = sim
        self.workload_random = workload_rand
        self.config = config
    
    def get_traffic_classes(self) -> List[List[int]]:
        return self.config['tClasses']
    
    def get_expected_workload_time(self) -> int:
        return self.config['workload_start_new_flows_till']
    
    def prepare_before_simulation_starts(self):
        raise NotImplementedError()
    
    def get_workload_statistics(self):
        raise NotImplementedError()
    
    def __next__(self):
        raise NotImplementedError()


class SyntheticWorkloadGenerator(WorkloadGenerator):
    
    @staticmethod
    def get_default_config():
        config = {
            'workload_start_new_flows_till': 10 * 1000000,
            
            # parameters for two level markov model
            'workload_probability_stay_in_l': 0.8,
            'workload_probability_stay_in_h': 0.4,
            # the workload generator takes this value and multiplies it with _l and _h to get the real probabilities
            'workload_probability_factor': 0.8,
            'workload_flow_arrival_l': 120,
            'workload_flow_arrival_h': 15,
            
            # lambda is used for: workload_flow_arrival_l * lambda and workload_flow_arrival_h * lambda
            'workload_lambda': 60,
            
            'workload_packet_inter_arrival_expected_time': 800,
            'workload_packets_per_flow': 150,
            
            'number_of_sf_types': 2,
            
            # processing rate is per 1s, when using one cpu share, so we normalze the rate to servers with a capacity of 80
            'sf_processing_rate': [
                1000000 // 160 // 80,
                1000000 // 250 // 80,
            ],
            
            # configure sfc traffic classes
            'tClasses': [[0, 1],
                         [0],
                         [1]],
            
            # we take the deadline of each sfc times "workload_deadline_scaling"
            'workload_deadline_scaling': 10
        }
        
        # we set the deadline of each sfc according to the delay each of its sf requires to process.
        config['workload_deadline_per_packet'] = []
        for tc in config['tClasses']:
            delay = 0
            for sf in tc:
                delay += 1000000 / config['sf_processing_rate'][sf]
            config['workload_deadline_per_packet'].append(int(delay))
        
        return config
    
    def __init__(self, sim: 'Sim', workload_rand: random.Random, config: Dict):
        super().__init__(sim, workload_rand, config)
        self.packet_generator = None
        self.flows = None
    
    def __next__(self):  # -> "IngressEvent"
        return self.packet_generator.__next__()
    
    def prepare_before_simulation_starts(self):
        self.packet_generator = self.create_packet_generator()
        
        if next(self.packet_generator) is not None:
            raise NameError("the very first call must be None")
    
    def get_workload_statistics(self):
        # what is the packets to be processed for each sf type
        flow_pros: Flow.Props = self.sim.props.flow
        flow_size = self.config['workload_packets_per_flow']
        # number of packets, p/s, req. processing time per second at with 1 server capacity (so this is also the req. capacity)
        per_sf_demand = {sf: [0, 0, 0] for sf in range(len(self.sim.props.sfi.processingRateOfSfType))}
        for flow in self.flows:
            index = 0
            eoc = None
            while eoc is None or not eoc:
                sf, eoc = flow_pros.sfc_class_to_sf[flow.sfc_class + index]
                per_sf_demand[sf][0] += flow_size
                index += 1
        
        exp_time = self.get_expected_workload_time() / 1000000  # in seconds
        sfi_props: SFI.Props = self.sim.props.sfi
        
        for sf in per_sf_demand:
            # per second, how many packets to serve
            per_sf_demand[sf][1] = per_sf_demand[sf][0] / exp_time
            # how much processing time does it take with one server capacity
            per_sf_demand[sf][2] = per_sf_demand[sf][1] / sfi_props.processingRateOfSfType[sf]
        
        return per_sf_demand
    
    def create_packet_generator(self) -> Generator:
        traffic_class = self.get_traffic_classes()
        workload_flow_arrival_l = int(self.config['workload_lambda'] * self.config['workload_flow_arrival_l'])
        workload_flow_arrival_h = int(self.config['workload_lambda'] * self.config['workload_flow_arrival_h'])
        workload_packet_inter_arrival_expected_time = self.config['workload_packet_inter_arrival_expected_time']
        workload_packets_per_flow = self.config['workload_packets_per_flow']
        workload_probability_stay_in_l = self.config['workload_probability_stay_in_l'] * self.config[
            'workload_probability_factor']
        workload_probability_stay_in_h = self.config['workload_probability_stay_in_h'] * self.config[
            'workload_probability_factor']
        workload_start_new_flows_till = self.config['workload_start_new_flows_till']
        
        workload_deadline_per_packet = [int(self.config['workload_deadline_scaling'] * d)
                                        for d
                                        in self.config['workload_deadline_per_packet']]
        
        assert len(workload_deadline_per_packet) == len(traffic_class)
        
        self.config['workload_effective_deadline_per_packet'] = workload_deadline_per_packet
        print(f"Deadlines randomized: {workload_deadline_per_packet}")
        
        burstiness = 1.0 / workload_probability_stay_in_h + 1.0 / workload_probability_stay_in_l
        print(f"Workload has a burstiness of {burstiness}")
        
        all_sff = list(self.sim.props.sff.allSFFs.values())
        flow_arrival_state_high = False
        self.flows: Deque[Flow] = deque()
        number_of_all_flows_hops = 0
        
        ingress_nodes = all_sff[:]
        
        for ingress in ingress_nodes:
            flow_start_time = 0
            
            while flow_start_time < workload_start_new_flows_till:
                # switch state of flow arrival times?
                if flow_arrival_state_high:
                    # transition to low state?
                    if np.random.rand() > workload_probability_stay_in_h:
                        flow_arrival_state_high = False
                else:
                    if np.random.rand() > workload_probability_stay_in_l:
                        flow_arrival_state_high = True
                
                next_inter_flow_time = np.random.poisson(workload_flow_arrival_h if flow_arrival_state_high
                                                         else workload_flow_arrival_l)
                flow_start_time += next_inter_flow_time
                
                egress = np.random.choice(all_sff)
                
                random_traffic_class_index = np.random.choice(range(len(traffic_class)))
                random_traffic_class = traffic_class[random_traffic_class_index]
                
                self.flows.append(Flow(sim=self.sim, sf_type_chain=random_traffic_class,
                                       qos_max_delay=workload_deadline_per_packet[random_traffic_class_index],
                                       desired_egress_ssf_id=egress.id,
                                       ingress_sff_id=ingress.id,
                                       start_time=flow_start_time))
                number_of_all_flows_hops += len(random_traffic_class)
        
        print(f"Created {len(self.flows)} flows.. start sorting now")
        # sort flows based on start_time
        self.flows = deque(sorted(self.flows, key=lambda f: f.start_time))
        
        flow_sizes = np.random.poisson(workload_packets_per_flow, len(self.flows)).tolist()
        print(f"All flows have in total {number_of_all_flows_hops} "
              f"hops, i.e., a scheduler has to take at least that many decisions")
        
        yield None
        
        while len(self.flows) > 0:
            f = self.flows.popleft()
            single_flow_size = flow_sizes.pop()
            packet_start_times = list(
                np.random.poisson(workload_packet_inter_arrival_expected_time, single_flow_size))
            
            for packet in range(single_flow_size):
                yield Packet.create_wrap_in_event(time_ingress=int(f.start_time + packet_start_times.pop()),
                                                  flow=f,
                                                  transmission_size=1)



#!/usr/bin/env python3
# coding=utf-8
from sortedcontainers import SortedKeyList


class BaseEvent(object):
    def __init__(self, at_time):  # absolute time of event (from start of simulation)
        # we don't allow events with floating time, since there might be a precision error -> may affect
        # deterministic property int, because float may introduce rounding errors and is more expensive
        self.time = int(at_time)
        self.ignoreWhenFinished = False
    
    def process_event(self):
        pass
    
    def __str__(self):
        return f"{self.__class__.__name__}({self.time})"
    
    def __repr__(self):
        return str(self)


class PacketHoldingEvent(BaseEvent):
    
    def __init__(self, at_time, inner_packet):
        super().__init__(at_time)
        self.inner_packet = inner_packet
    
    def update_packet_time_tracking(self):
        raise NameError("this need to be implemented")


class EventList(object):
    length_of_a_slice = 800000
    
    def __init__(self, sim):
        self.current_list: SortedKeyList[BaseEvent] = SortedKeyList(key=lambda e: e.time)
        self.sim = sim
        self.number_of_events = 0
        self.number_of_relevant_events = 0
        self.last_popped_time = None
        self.last_time_of_relevant_event = None
    
    def debug_get_all_remaining(self):
        return self.current_list[:]
    
    def debug_access(self, relative_to_next_event) -> BaseEvent:
        assert relative_to_next_event >= 0
        return self.current_list[relative_to_next_event]
    
    # this is used when filling up the backlog with new events from the workload
    def add_event_from_the_back(self, event):
        self.enqueue_event(event)
    
    def enqueue_event(self, event: BaseEvent):
        self.number_of_events += 1
        if not event.ignoreWhenFinished:
            self.number_of_relevant_events += 1
            if self.last_time_of_relevant_event is None or self.last_time_of_relevant_event < event.time:
                self.last_time_of_relevant_event = event.time
        
        self.current_list.add(event)
    
    def len(self):
        return self.number_of_events
    
    def get_number_of_relevant_events(self):
        return self.number_of_relevant_events
    
    def pop_next(self) -> BaseEvent:
        item = self.current_list.pop(0)
        
        self.number_of_events -= 1
        if not item.ignoreWhenFinished:
            self.number_of_relevant_events -= 1
        assert self.last_popped_time is None or item.time >= self.last_popped_time
        self.last_popped_time = item.time
        
        if not self.sim.packed_generator_is_done and self.last_time_of_relevant_event - 500000 < self.last_popped_time:
            self.sim.ask_packet_generator_for_more_events(EventList.length_of_a_slice, 800000)
        
        return item
    
    def print_snapshot(self):
        print(f"Number of pending events: {self.number_of_events}")
        if len(self.current_list) > 0:
            for e in self.current_list[:5]:
                print(f"    {e}, {e.time}")
            print("...")
            for e in self.current_list[-5:]:
                print(f"    {e}, {e.time}")
    
    def peek_last(self):
        return self.current_list[-1]



#!/usr/bin/env python3
# coding=utf-8

import datetime
import gc
import pprint
import random
import re
import time
import traceback
from typing import Generator

import numpy as np

from . import __version__, __copyright__
from .events import EventList, BaseEvent, PacketHoldingEvent


class SchedulingFailure(Exception):
    def __init__(self, message):
        super().__init__(message)


class Sim(object):
    # hooks
    reset_global_fields = []
    simulation_done_hooks = []
    simulation_start_hooks = []
    stop_sim_catch_all_packets_hooks = []
    simulation_one_liner_hooks = []
    workload_over_hooks = []
    ui_update_hook = None
    
    @classmethod
    def register_reset_global_fields(cls, clazz):
        cls.reset_global_fields.append(clazz)
        print(f'::register class {clazz}')
        return clazz
    
    @classmethod
    def register_simulation_done_hook(cls, func):
        cls.simulation_done_hooks.append(func)
        return func
    
    @classmethod
    def register_workload_over_hook(cls, func):
        cls.workload_over_hooks.append(func)
        return func
    
    @classmethod
    def register_simulation_start_hook(cls, func):
        cls.simulation_start_hooks.append(func)
        return func
    
    @classmethod
    def register_sim_oneliner_text_provider(cls, func):
        cls.simulation_one_liner_hooks.append(func)
        return func
    
    # these hooks provide iterator access to all packets that are somewhere stored at internal data structures
    @classmethod
    def register_stop_sim_catch_all_packets_hooks(cls, func):
        cls.stop_sim_catch_all_packets_hooks.append(func)
        return func
    
    @classmethod
    def register_ui_update_hook(cls, f):
        if cls.ui_update_hook is not None:
            print('WARNING someone overwrites ui update hook')
        cls.ui_update_hook = f
        return f
    
    def __init__(self, seed):
        self.packet_generator: Generator = None
        self.workload_end_time: int = 0
        self.last_packet_ingress_time = 0
        self.ui_shows_full_state = True
        self.start_time = None
        self.stats = None
        self.current_event = None
        self.PACKET_ID_TO_DEBUG: int = None
        self.STOP_SIMULATION_IF_SCHEDULER_WAS_UNSUCCESSFUL: bool = False
        self.TRACE_PACKET_PATH: bool = False
        self.KEEP_LIST_OF_ALL_PACKETS: bool = False
        self.SERVER_CPU_POLICY_DYNAMIC_INTERVAL: int = 1000000
        self.DEBUG: bool = False
        self.SERVER_CPU_SHARE_GRANULARITY: int = 10000
        self.run: bool = False
        self.currentTime: int = 0
        self.event_list: EventList = EventList(sim=self)
        self.ignore_all_future_schedule_event_attempts: bool = False
        self.lastRelevantTime = 0
        self.printing_progress_previous_ticks = 0
        self.printing_progress_previous_time = 0
        self.random = random.Random(seed)
        self.random_state_workload_python = None
        self.random_state_workload_numpy = None
        self.packed_generator_is_done = False
        
        print(f"# SFC TSS - Traffic Scheduling Simulator")
        print(f"# Version {__version__}")
        
        # initialize global fields
        class Props:
            def __init__(self):
                for property_class in Sim.reset_global_fields:
                    properties = property_class()
                    try:
                        name = re.findall(r'([a-zA-Z_][a-zA-Z0-9_]*)', properties.__class__.__qualname__)[-2]
                        name_of_props = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name).lower()
                    except IndexError as e:
                        print(f'error while resetting properties for {property_class}, '
                              f'{e}: {traceback.format_exc()}')
                        raise e
                    # print(f'::reset properties for "sim.{name_of_props}"')
                    setattr(self, name_of_props, properties)
            
            def __str__(self):
                return pprint.pformat(vars(self))
        
        self.props = Props()
    
    def schedule_event(self, event: BaseEvent):  # schedules an event -> simply add it to the list of events
        if self.ignore_all_future_schedule_event_attempts:
            return
        self.event_list.enqueue_event(event)
    
    def debug_get_sorted_event_list(self):
        return sorted(self.event_list.debug_get_all_remaining(), key=lambda x: x.time)
    
    def calculate_simple_statistics(self):
        successfully_delivered = (
                self.props.packet.statsPacketsSuccessfulProcessed - self.props.packet.counter_packet_after_workload_end_in_system_no_timeout)
        idle_time = 0
        for s in self.props.server.all_servers:
            idle_time += s.stats_idle_time
            if s.is_free():
                idle_time += self.currentTime - s.stats_last_time_idle
        idle_time_ratio = 0 if self.currentTime == 0 else round(
            100 * (idle_time / len(self.props.server.all_servers)) / self.currentTime, 4)
        
        self.stats = {'total_scheduling_attempts':
                          sum([s.scheduler.scheduling_attempts for s in self.props.sff.allSFFs.values()]),
                      'success_rate': self.props.packet.statsPacketsSuccessfulProcessed /
                                      self.props.packet.statsPacketsTotalCount,
                      'reject_rate': self.props.packet.statsPacketsRejectedSchedule /
                                     self.props.packet.statsPacketsTotalCount,
                      'service_quality': 0.0 if successfully_delivered == 0 else
                      1 - self.props.packet.statsRatiosQos / successfully_delivered,
                      'server_idle_time': idle_time,
                      'server_idle_time_ratio': idle_time_ratio
                      }
    
    def print_sim_snapshot(self, per_sfi_stats: bool = False):
        from sfctss import sanity
        print("-" * 20)
        print(f"Sim Snapshot at time:{self.currentTime}µs ({datetime.datetime.now()})")
        sanity.print_sim_snapshot(self, per_sfi_stats=per_sfi_stats)
        print("-" * 20)
        self.event_list.print_snapshot()
        print('#' * 20)
    
    def run_one_step(self, print_status=False, do_not_stop=False):
        if not self.run:
            self.run = True
        
        if self.event_list.len() > 0:
            e = self.event_list.pop_next()
            while (e.ignoreWhenFinished and not do_not_stop) and self.event_list.len() > 0:
                e = self.event_list.pop_next()
            self.current_event = e
            if e.ignoreWhenFinished and not do_not_stop:
                return False
            assert e.time >= self.currentTime
            self.currentTime = e.time  # update time to the time of the current event
            if print_status:
                print(f'-- @{self.currentTime}')
            e.process_event()
            return True
        else:
            return False
    
    def run_sim(self, max_sim_time: int = -1, show_progress: bool = True,
                interactive=False, ui=False,
                stop_simulation_when_workload_is_over=False):  # method to run the simulator
        assert not self.run
        assert not (ui and show_progress)
        if self.packet_generator is None:
            self.packed_generator_is_done = True
        self.run = True
        # we need to check this, in case we have only one SFF, so there is no
        # link setup
        for f in self.simulation_start_hooks:
            f(self)
        ticks = 0
        self.start_time = time.time()
        peek_last = self.event_list.peek_last()
        if self.workload_end_time == 0:
            self.workload_end_time = peek_last.time
        self.event_list.print_snapshot()
        print("start simulation @{0}µs till max SimTime:{1}".format(
            str(self.start_time), str(max_sim_time)))
        
        still_during_workload = True
        
        if self.currentTime != 0:
            raise NameError("this Implementation uses many class instances, \
            hence, your cannot run multiple simulations w/o restarting python")
        
        if ui:
            from . import ui
        
        try:
            # main loop
            
            while self.event_list.get_number_of_relevant_events() > 0 and (
                    max_sim_time == -1 or max_sim_time >= self.currentTime):
                ticks += 1  # count ticks for statistics
                if show_progress:
                    if ticks % 4096 == 0:
                        self.update_sim_status_oneliner(ticks=ticks)
                if ui:
                    if ticks % 4096 == 0:
                        self.update_ui(ticks)
                        if interactive:
                            input("#" * 40 + "\nPress Enter to continue...")
                
                e = self.event_list.pop_next()
                assert e != self.current_event
                
                if e.time > self.workload_end_time and still_during_workload and stop_simulation_when_workload_is_over:
                    # put the event back
                    self.event_list.enqueue_event(e)
                    
                    # we are now done with the workload
                    # inform all statistic writer that workload is over
                    self.ignore_all_future_schedule_event_attempts = True
                    
                    # check all packets which are still in the system,
                    # and for each, check the current time in system,
                    # if higher than deadline, consider as drop,
                    #   drop, we fire p.drop, so that we get all statistics
                    # if lower, consider as successful delivery
                    #   no drop, but we cannot account into the detailed statistics.
                    #   so we simply decrease the number of packets in system,
                    #   increase successful by 1, but we do not fire
                    #   p.done, so that the avg counter are not affected
                    
                    def check_remaining_packets():
                        for callback in self.stop_sim_catch_all_packets_hooks:
                            for packet in callback(self):
                                yield packet
                        
                        while self.event_list.len() > 0:
                            event = self.event_list.pop_next()
                            if isinstance(event, PacketHoldingEvent):
                                event.update_packet_time_tracking()
                                yield event.inner_packet
                    
                    p_finder = check_remaining_packets()
                    for p in p_finder:
                        p.handle_stop_simulation()
                    
                    for f in self.workload_over_hooks:
                        f(self)
                    
                    self.update_sim_status_oneliner(ticks, newline=True)
                    print(
                        "simulator reaches end of workload @{0}µs stop simulation ".format(
                            self.currentTime))
                    
                    break
                
                self.current_event = e
                self.currentTime = e.time  # update time to the time of the current event
                if not e.ignoreWhenFinished:
                    self.lastRelevantTime = self.currentTime
                
                if self.DEBUG:
                    print(f"Simulation tick @time {self.currentTime} with event {e}")
                
                e.process_event()  # process this event, regardless of which event it is
                
                if self.event_list.get_number_of_relevant_events() == 0:
                    self.update_sim_status_oneliner(ticks, newline=True)
                    print(
                        "simulator is done @{0}µs because of empty event stack of relevant events ".format(
                            self.currentTime))
                    
                    break
                
                if interactive and not ui:
                    self.print_sim_snapshot()
                    input("#" * 40 + "\nPress Enter to continue...")
            
            self.ignore_all_future_schedule_event_attempts = True
            
            if self.event_list.get_number_of_relevant_events() > 0:
                print(f"\nsimulator stops @{self.currentTime} because of simulation " +
                      f"reached given max sim time {max_sim_time}")
            
            # process all events which are still in the list of events, but do not belong to relevant events (e.g.
            # statistics)
            if self.event_list.len() > 0 and self.event_list.get_number_of_relevant_events() == 0:
                print(f"Finally process all pending statistic events")
                while self.event_list.len() > 0:
                    self.event_list.pop_next().process_event()
            
            for f in self.simulation_done_hooks:
                f(self)
        
        except BaseException as e:
            print(
                "#" *
                40 +
                "\n" +
                "#" *
                40 +
                f"\nsomething went wrong @{self.currentTime}, {self.event_list.len()} events "
                f"try to write all statistics to disk\n")
            self.print_sim_snapshot()
            raise e
        finally:
            pass
        
        end_time = time.time()
        time_passed = (end_time - self.start_time)
        performance = ticks / time_passed
        print("Simulation done with a total of {0} ticks, ~{1} ticks/s, {2}x speedup ({3}s simulated in {4}s)".format(
            str(ticks), str(int(round(performance, 0))), str(
                round(self.lastRelevantTime / time_passed / 1000000, 2)),
            str(self.lastRelevantTime / 1000000), str(time_passed)))
        
        self.calculate_simple_statistics()
        print(self.stats)
        
        # clear events that are not important anymore
        # e.g., a statistic event might still be scheduled
        # we have to remove these events to check if there are some "packet"
        # events in the queue
        remaining_events = [e for e in self.event_list.debug_get_all_remaining() if not e.ignoreWhenFinished]
        
        if len(remaining_events) > 0:
            if stop_simulation_when_workload_is_over:
                print(f"simulation has still {self.event_list.get_number_of_relevant_events()} enqueued events," +
                      f" but we stopped the simulation")
            else:
                print("simulation has still {0} enqueued events".format(
                    str(len(remaining_events))))
    
    def update_sim_status_oneliner(self, ticks, newline=False):
        progress = 100 * self.currentTime // self.workload_end_time \
            if self.currentTime <= self.workload_end_time else -1
        time_passed = round(time.time() - self.start_time, 1) + 0.1
        performance = ticks // time_passed
        
        performance_recent = int((ticks - self.printing_progress_previous_ticks) //
                                 (time.time() - self.printing_progress_previous_time))
        
        bar = (10 - int(progress / 10)) if progress >= 0 else 0
        progress_text = f"{progress}%" if progress >= 0 else "workload finished"
        print(f"\r[{'#' * (10 - bar) + ' ' * bar}] {progress_text} "
              f"@{self.currentTime}µs/"
              f"{self.event_list.peek_last().time if self.event_list.get_number_of_relevant_events() > 0 else '/'} "
              f"({performance_recent} t/s avg: {performance} t/s in {int(time_passed)} s) "
              f"{' '.join([provider(self) for provider in self.simulation_one_liner_hooks])}",
              end="\n" if newline else '\r')
        self.printing_progress_previous_ticks = ticks
        self.printing_progress_previous_time = time.time()
    
    def update_ui(self, ticks, split_server_table=True):
        if self.currentTime == 0:
            return
        Sim.ui_update_hook(self, ticks, split_server_table)
    
    def register_packet_generator(self, packet_generator, fetch_all=False):
        assert self.packet_generator is None
        self.packet_generator = packet_generator
        
        self.workload_end_time = packet_generator.get_expected_workload_time()
        try:
            print(f"Ask packet generator for a slice of up to {EventList.length_of_a_slice} packets")
            for _ in range(EventList.length_of_a_slice):
                self.event_list.add_event_from_the_back(self.packet_generator.__next__())
            
            while fetch_all:
                self.event_list.add_event_from_the_back(self.packet_generator.__next__())
            
            # freezing random states when done
            self.random_state_workload_python = random.getstate()
            self.random_state_workload_numpy = np.random.get_state()
        except StopIteration:
            self.packed_generator_is_done = True
            print("packet gen is done")
        self.workload_end_time = max(self.workload_end_time, self.event_list.last_time_of_relevant_event)
    
    def ask_packet_generator_for_more_events(self, minimum_number_of_new_events: int,
                                             minimum_future_time: int):
        
        if not self.packed_generator_is_done and self.packet_generator is not None:
            try:
                random.setstate(self.random_state_workload_python)
                np.random.set_state(self.random_state_workload_numpy)
                
                print(f"\rloading @{self.currentTime} ", end="\r")
                
                gc.collect()
                
                time_before = self.event_list.last_time_of_relevant_event
                for _ in range(minimum_number_of_new_events):
                    self.event_list.add_event_from_the_back(self.packet_generator.__next__())
                expected_time = time_before + minimum_future_time
                while self.event_list.last_time_of_relevant_event < expected_time:
                    self.event_list.add_event_from_the_back(self.packet_generator.__next__())
            
            except StopIteration:
                self.packed_generator_is_done = True
                # print("packet gen is done")
            finally:
                self.random_state_workload_python = random.getstate()
                self.random_state_workload_numpy = np.random.get_state()
                
                self.workload_end_time = max(self.workload_end_time, self.event_list.last_time_of_relevant_event)



#!/usr/bin/env python3
# coding=utf-8
import math
import os

from .simulator import Sim


@Sim.register_ui_update_hook
def update_ui(sim: Sim, ticks, split_server_table):
    fill_packet = len(str(sim.props.packet.lastId))
    fill_sfi = len(str(sim.props.sfi.next_free_id))
    fill_sff = len(str(len(sim.props.sff.allSFFs)))
    fill_sf = len(str(len(sim.props.sfi.processingRateOfSfType)))
    fill_server = len(str(sim.props.server.lastId))
    os.system('cls' if os.name == 'nt' else 'clear')
    print("#" * 50)
    sim.update_sim_status_oneliner(ticks, newline=True)
    print("#" * 50)
    
    left_margin = 34 + fill_sff + fill_packet
    
    packets_seen = sim.props.packet.counter_packet_in_system + sim.props.packet.statsPacketsSuccessfulProcessed + \
                   sim.props.packet.statsPacketsRejectedProcessingDelay + sim.props.packet.statsPacketsRejectedSchedule
    print(f' packets in the system: {str(sim.props.packet.counter_packet_in_system).rjust(fill_packet, " ")} ' +
          f' total SFF-SFF hops: {sim.props.sff.counter_packets_put_on_wire} '
          f'({round(sim.props.sff.counter_packets_put_on_wire / packets_seen, 2)} hops / packet)')
    print("#" * 50)
    alphas_per_class = dict()
    for queue in range(sim.props.flow.sfc_next_free_class):
        alpha_normalized_enumerator = math.pow(sim.props.flow.max_deadline, 2)
        alpha_denominator = math.pow(sim.props.flow.sfc_class_to_deadline[queue], 2)
        alphas_per_class[queue] = alpha_normalized_enumerator / alpha_denominator
    
    print(
        f"{' alpha values per class '.rjust(left_margin, ' ')}"
        f"{[str(round(x, 2)).rjust(fill_packet, ' ') for queue in sorted(alphas_per_class.keys()) for x in [alphas_per_class[queue]]]}")
    out = [(queue, f"{(str(sf) + ' ' + ('EX ' if last else '  ')).rjust(fill_packet, ' ')}") for queue, (sf, last) in
           sim.props.flow.sfc_class_to_sf.items()]
    out = sorted(out, key=lambda x: x[0])
    
    print(f"{' sf type of class, and last_of_sfc flag '.rjust(left_margin, ' ')}{[x for _, x in out]}")
    print(
        f"{' deadline of the class '.rjust(left_margin, ' ')}"
        f"{[str(sim.props.flow.sfc_class_to_deadline[x]).rjust(fill_packet, ' ') for x in range(sim.props.flow.sfc_next_free_class)]}")
    
    print(" service rate per sf: " + str(
        [f"SF {i}: {(round(sim.props.sfi.processingRateOfSfType[i], 1))}" for i in
         range(len(sim.props.sfi.processingRateOfSfType))]))
    
    print("#" * 50)
    print("# SFF Overview: SFF.id | total items in queue | items per packet class")
    
    for sff in sim.props.sff.allSFFs.values():
        queues = ""
        
        if sff.scheduler.requires_queues_per_class():
            queues = [str(len(sff.packet_queue_per_class[p_class])).rjust(fill_packet, '_') for p_class in
                      range(sim.props.flow.sfc_next_free_class)]
        
        raw_rates = [round(sff.scheduler.get_arrival_rate_estimate(p_class))
                     for p_class in range(len(sim.props.sfi.processingRateOfSfType))]
        raw_load = [
            f'{round(raw_rates[p_class] / sff.scheduler.mySFF.service_rate_per_sf[p_class], 2) if p_class in sff.scheduler.mySFF.service_rate_per_sf else "/"}|{raw_rates[p_class]}'
            for p_class in range(len(sim.props.sfi.processingRateOfSfType))]
        
        print(f" ID{str(sff.id).rjust(fill_sff, '_')}:  " +
              f"{str(sff.get_number_of_queued_packets()).rjust(fill_packet, '_')} " +
              f"packets in queue, classes: {queues}")
        
        print(f"rate|load /sf; total: {str(raw_load).rjust(fill_packet, ' ')}: ".rjust(left_margin, ' '))
    
    # print the drops per class
    drops = [str(sim.props.flow.statistics_drops_per_class[x]).rjust(fill_packet, '_') for x in
             range(sim.props.flow.sfc_next_free_class)]
    print("")
    print("drops per class: ".rjust(left_margin, ' ') + str(drops))
    
    if not sim.ui_shows_full_state:
        return
    
    print("#" * 50)
    print("# SFI Overview ")
    
    per_sf_queue = {sf: 0 for sf in range(len(sim.props.sfi.processingRateOfSfType))}
    per_sf_active = {sf: 0 for sf in range(len(sim.props.sfi.processingRateOfSfType))}
    
    per_sf_distinct_server = {sf: set() for sf in range(len(sim.props.sfi.processingRateOfSfType))}
    for sfi in sim.props.sfi.all_sfi.values():
        if not sfi.free:
            per_sf_active[sfi.of_type] += 1
        per_sf_distinct_server[sfi.of_type].add(sfi.server)
        per_sf_queue[sfi.of_type] += len(sfi.queue)
    
    for sf in per_sf_queue:
        print(
            f" SF{str(sf).rjust(fill_sf, '0')}: {str(per_sf_queue[sf]).rjust(fill_packet, '_')} packets in queue, " +
            f" [{str(per_sf_active[sf] * '#').ljust(len(per_sf_distinct_server[sf]), '_')}] " +
            f"active SFIs (out of {len([x for x in sim.props.sfi.all_sfi.values() if x.of_type == sf])})")
    
    print("#" * 50)
    print("# SFI/Server details: for each server, id | capacity | idle time")
    server_messages = []
    
    for server in sim.props.server.all_servers:
        idle_time = server.stats_idle_time
        if server.is_free():
            idle_time += sim.currentTime - server.stats_last_time_idle
        
        server_messages.append(f" Server {str(server.id).rjust(fill_server, '_')}, " +
                               f"{server.processing_cap}units | {str(round(idle_time / sim.currentTime * 100, 1)).rjust(4, ' ')}% idle")
        for sfi in server.SFIs:
            server_messages.append(f"   SFI {str(sfi.id).rjust(fill_sfi, '_')} of type " +
                                   f"{str(sfi.of_type).rjust(fill_sf, '_')}:  " +
                                   f"{'OFF' if sfi.free else 'ON '} " +
                                   f" {str(len(sfi.queue)).rjust(fill_packet, '_')} packets in queue")
    
    fill_line = max([len(str(l)) for l in server_messages]) + 2
    
    if split_server_table:
        left = server_messages[:len(server_messages) // 2]
        right = server_messages[len(left):]
        
        while len(left) > 0 and len(right) > 0:
            print(
                f'{left.pop(0)}'.ljust(fill_line, ' ') + (' ' if len(left) % 2 == 0 else '|') + f"  {right.pop(0)}")
        print((left + right + [""]).pop(0))
    else:
        for m in server_messages:
            print(m)



# coding=utf-8

from .core import *
from collections import deque
from typing import Set, Tuple
import math



class GreedyShortestDeadlineFirstScheduler(BaseScheduler):
    
    def __init__(self, sim: Sim,
                 incremental: bool = True, oracle=True,
                 admission_control_threshold_low: float = 0.8,
                 admission_control_threshold_high: float = 1.1
                 ):
        super().__init__(sim=sim,
                         incremental=incremental,
                         oracle=oracle,
                         activate_acp=not oracle,
                         admission_control_threshold_low=admission_control_threshold_low,
                         admission_control_threshold_high=admission_control_threshold_high)
        self.admission_control_threshold_low = admission_control_threshold_low
        self.admission_control_threshold_high = admission_control_threshold_high
        
        assert admission_control_threshold_low < admission_control_threshold_high
        
        if not self.oracle:
            self.rate_estimator: Dict[int, RateEstimator] = {}
    
    def is_always_able_to_build_full_path(self):
        return self.oracle and not self.incremental
    
    def apply_scheduling_logic_for_packet(self, packet: Packet):
        sff_props: SFF.Props = self.sim.props.sff
        
        # get the packet from the scheduler's queue
        if self.requires_queues_per_class():
            popped_packet = self.mySFF.packet_queue_per_class[Flow.get_packet_class_of_packet(packet)].pop()
        else:
            popped_packet = self.mySFF.packet_queue.pop()
        
        assert popped_packet == packet
        
        packet.timeQueueScheduling += packet.get_delta_of_time_mark()
        
        self.mark_time_scheduling_starts()
        
        # this packet starts from this SFF, so search from this point
        p_at_sff: SFF = self.mySFF
        
        if self.sim.DEBUG:
            print(". heuristic scheduler @ {0}".format(p_at_sff))
        
        scheduled_path = []
        
        incremental_path = False
        
        # for each remaining sf type of the requested chain of this packet
        while 0 < len(packet.toBeVisited) and not incremental_path:
            next_sf_type = packet.toBeVisited.pop(0)
            # we need to get a SFI for this corresponding sf type
            # try to get this SFI from the SFF where this packet is currently
            # (p_at_sff)
            
            sff_to_check = sff_props.allSFFs if self.oracle else [p_at_sff.id]
            
            sfi_to_check = []
            #  get a list of all possible SFIs
            # calculate for each of these SFIs the cost value
            best_latency = -1
            for sffIDToAsk in sff_to_check:
                sff_to_ask = sff_props.allSFFs[sffIDToAsk]
                delay_of_sff_connection = 0 if p_at_sff == sff_to_ask else SFF.get_multi_hop_latency_for(self.sim,
                                                                                                         p_at_sff.id,
                                                                                                         sff_to_ask.id)
                if best_latency != -1 and best_latency <= delay_of_sff_connection:
                    # we skip this sff if the latency to this sff is bigger than the best option we found so far
                    continue
                
                if next_sf_type in sff_to_ask.SFIsPerType:
                    for sfi in sff_to_ask.SFIsPerType[next_sf_type]:
                        delay = sfi.get_expected_waiting_time() + sfi.get_expected_processing_time()
                        if p_at_sff != sff_to_ask:
                            delay += SFF.get_delay_of_connection(p_at_sff, sff_to_ask)
                        if best_latency == -1:
                            best_latency = delay
                        best_latency = min(best_latency, delay)
                        sfi_to_check.append((delay, sff_to_ask, sfi))
            
            if len(sfi_to_check) == 0:
                if not self.oracle:
                    raise NameError(f'something is going wrong. I don\'t have any SFI which could serve this packet,'
                                    f'but ACP should have handeled this case!?')
                
                packet.realTimeScheduling += self.get_time_delta_of_scheduling()
                try:
                    raise SchedulingFailure("failed to find a path for {0}".format(
                        str(packet.flow.sfTypeChain)))
                except SchedulingFailure as e:
                    raise e
                finally:
                    packet.reject()
            
            # find the sfi with the lowest delay till packet is processed
            best_sfi = sfi_to_check[0]
            for sfi_tuple in sfi_to_check:
                if sfi_tuple[0] <= best_sfi[0]:
                    # update the best sfi, but for equal values we prefer to stay at our sff
                    if sfi_tuple[0] != best_sfi[0] or sfi_tuple[1] == self.mySFF:
                        best_sfi = sfi_tuple
            
            sff_to_ask = best_sfi[1]
            sfi = best_sfi[2]
            
            if self.sim.DEBUG:
                print(". found a SFI at SFF {0}".format(sff_to_ask))
            
            # do we need to go to another SFF, or do we stay at the
            # current SFF?
            if p_at_sff != sff_to_ask:
                # we need to go to a different SFF, but before,
                # we have to go back to the original SFF, so that
                # if the packet is at a SFI, it goes fist back
                # to the SFF of this SFI, and then to the next SFF
                
                # so go to p_at_sff if the previous path element was
                # a SFI
                if (len(scheduled_path) >
                        0 and scheduled_path[-1][0] == SFI.__name__):
                    scheduled_path.append((SFF.__name__, p_at_sff))
                
                path_to_other_sff = SFF.get_multi_hop_path_for(self.sim, p_at_sff.id, sff_to_ask.id)
                if self.sim.DEBUG:
                    print(". path to this guy contains {0} intermediate SFFs".format(len(path_to_other_sff)))
                for next_sff in path_to_other_sff:
                    scheduled_path.append((SFF.__name__, sff_props.allSFFs[next_sff]))
                
                p_at_sff = sff_to_ask
            
            scheduled_path.append((SFI.__name__, sfi))
            
            # if we are in incremental scheduling mode, we set the incremental_path flag,
            # so that we stop scheduling from here on
            if self.incremental:
                # incremental_path, so we found a different SFF,
                # hence we stop scheduling here
                incremental_path = True
                # and then go back to the SFF for scheduling
                scheduled_path.append((SFF.__name__, sff_to_ask))
        
        if not incremental_path:
            # finally, add the egress SFF
            if p_at_sff.id != packet.flow.desiredEgressSSFid:
                # go back to the sff
                scheduled_path.append((SFF.__name__, p_at_sff))
                path_to_dest = SFF.get_multi_hop_path_for(self.sim, p_at_sff.id, packet.flow.desiredEgressSSFid)
                for sff_id in path_to_dest:
                    scheduled_path.append((SFF.__name__, sff_props.allSFFs[sff_id]))
            else:
                scheduled_path.append((SFF.__name__, sff_props.allSFFs[packet.flow.desiredEgressSSFid]))
        
        if self.sim.DEBUG:
            Packet.debug_print_path(scheduled_path)
        
        packet.fullPath += scheduled_path
        self.scheduling_attempts += 1
        
        if packet.id == self.sim.PACKET_ID_TO_DEBUG:
            print("** debug packet visited the scheduler, current status:")
            Packet.debug_packet(packet)
        
        packet.realTimeScheduling += self.get_time_delta_of_scheduling()
        self.mySFF.handle_packet_from_scheduler(packet)


class LoadUnawareRoundRobinScheduler(BaseScheduler):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.sfi_round_robin_marker = dict()
    
    def __init__(self, sim: Sim, incremental: bool = True, oracle=True):
        super().__init__(sim=sim,
                         incremental=incremental,
                         oracle=oracle,
                         activate_acp=False)
    
    def applies_round_robin(self):
        return True
    
    def is_always_able_to_build_full_path(self):
        return not self.incremental
    
    def apply_scheduling_logic_for_packet(self, packet: Packet):
        sfi_props: SFI.Props = self.sim.props.sfi
        sff_props: SFF.Props = self.sim.props.sff
        
        # get the packet from the scheduler's queue
        if self.requires_queues_per_class():
            popped_packet = self.mySFF.packet_queue_per_class[Flow.get_packet_class_of_packet(packet)].pop()
        else:
            popped_packet = self.mySFF.packet_queue.pop()
        
        assert popped_packet == packet
        
        packet.timeQueueScheduling += packet.get_delta_of_time_mark()
        
        self.mark_time_scheduling_starts()
        
        # this packet starts from this SFF, so search from this point
        p_at_sff_id = self.mySFF.id
        
        scheduled_path = []
        
        # for each remaining sf type of the requested chain of this packet
        while 0 < len(packet.toBeVisited):
            next_sf_type = packet.toBeVisited.pop(0)
            
            # get the sfi which has to serve this packet
            if self.static_sfi_rates_per_sf_cum_weights is None:
                self.update_cum_weights_of_sfi_rates(include_own_sff=True)
            
            assert isinstance(next_sf_type, int)
            if (next_sf_type not in self.static_sfi_rates_per_sf_sorted_sfi
                    or next_sf_type not in self.static_sfi_rates_per_sf_cum_weights
                    or len(self.static_sfi_rates_per_sf_sorted_sfi[next_sf_type]) == 0
                    or len(self.static_sfi_rates_per_sf_cum_weights[next_sf_type]) == 0):
                packet.realTimeScheduling += self.get_time_delta_of_scheduling()
                packet.reject()
                if self.sim.DEBUG:
                    print(f"I'm not aware of any SFI of the required type:{next_sf_type}, so I have to reject the packet")
                raise SchedulingFailure(f"failed to find a path for {packet}")
            
            elif (len(self.static_sfi_rates_per_sf_sorted_sfi[next_sf_type]) !=
                  len(self.static_sfi_rates_per_sf_cum_weights[next_sf_type])):
                raise NameError(f'data structures broken')
            
            target_sfi_id = self.sim.random.choices(self.static_sfi_rates_per_sf_sorted_sfi[next_sf_type],
                                                    cum_weights=self.static_sfi_rates_per_sf_cum_weights[next_sf_type],
                                                    k=1)[0]
            
            target_sff_id = sfi_props.all_sfi[target_sfi_id].sffId
            
            # is this a different sff of where we are currently?
            
            if p_at_sff_id != target_sff_id:
                # we need to go to a different SFF, but before,
                # we have to go back to the original SFF, so that
                # if the packet is at a SFI, it goes fist back
                # to the SFF of this SFI, and then to the next SFF
                # (we need to check this, because thi scheduler supports to schedule the whole path at once)
                
                # so go to p_at_sff if the previous path element was
                # a SFI
                if (len(scheduled_path) >
                        0 and scheduled_path[-1][0] == SFI.__name__):
                    scheduled_path.append((SFF.__name__, sff_props.allSFFs[p_at_sff_id]))
                
                path_to_other_sff = SFF.get_multi_hop_path_for(self.sim, p_at_sff_id, target_sff_id)
                for p in path_to_other_sff:
                    scheduled_path.append((SFF.__name__, sff_props.allSFFs[p]))
                
                p_at_sff_id = target_sff_id
            
            scheduled_path.append((SFI.__name__, sfi_props.all_sfi[target_sfi_id]))
            
            if self.incremental:
                scheduled_path.append((SFF.__name__, sff_props.allSFFs[target_sff_id]))
                break
        
        if self.sim.DEBUG:
            Packet.debug_print_path(scheduled_path)
        
        packet.fullPath += scheduled_path
        self.scheduling_attempts += 1
        
        if packet.id == self.sim.PACKET_ID_TO_DEBUG:
            print("** debug packet visited the scheduler, current status:")
            Packet.debug_packet(packet)
        
        packet.realTimeScheduling += self.get_time_delta_of_scheduling()
        self.mySFF.handle_packet_from_scheduler(packet)


class MppScheduler(BaseScheduler):
    @Sim.register_reset_global_fields
    class Props:
        def __init__(self):
            self.allow_up_to_x_packets_underway_per_server: int = None
            self.do_sanity_checks: bool = False
            self.map_server_to_classes: dict = None
            self.r_matrix = None
            self.batch_scheduling = None
            self.blocked_sfi: Set[SFI] = set()
            self.packet_underway_counter_per_server: Dict[Server, int] = None
    
    def __init__(self,
                 sim: Sim,
                 incremental: bool = True, oracle=True,
                 block_sfi_while_packet_on_wire: bool = False,
                 consider_alpha_by_using_timeouts: bool = True,
                 allow_up_to_x_packets_underway_per_server: int = 1,
                 admission_control_threshold_low: float = 0.1,
                 admission_control_threshold_high: float = 1.3,
                 batch_scheduling: int = 1):
        super().__init__(sim=sim,
                         incremental=incremental,
                         oracle=oracle,
                         activate_acp=not oracle,
                         admission_control_threshold_high=admission_control_threshold_high,
                         admission_control_threshold_low=admission_control_threshold_low)
        
        self.block_sfi_while_packet_on_wire = block_sfi_while_packet_on_wire
        self.consider_alpha_by_using_timeouts = consider_alpha_by_using_timeouts
        self.free_server_count = -1  # important that we init with -1
        self.sim.props.mpp_scheduler.allow_up_to_x_packets_underway_per_server = allow_up_to_x_packets_underway_per_server
        
        self.assert_on_reject = False
        
        self.accessible_sf = None
        
        if self.sim.props.mpp_scheduler.batch_scheduling is None:
            self.sim.props.mpp_scheduler.batch_scheduling = batch_scheduling
        else:
            assert self.sim.props.mpp_scheduler.batch_scheduling == batch_scheduling
        
        if allow_up_to_x_packets_underway_per_server < batch_scheduling:
            raise NameError(f"Invalid configuration: batch_scheduling ({batch_scheduling}) > "
                            f"allow_underway_per_server ({allow_up_to_x_packets_underway_per_server})")
        
        if allow_up_to_x_packets_underway_per_server < 1:
            raise NameError("Invalid configuration, at least 1 packets needs to be underway")
        
        if not self.incremental:
            raise NameError("%s Scheduler does support incremental scheduling")
    
    def supports_cpu_policy(self, cpu_policy: ServerCpuPolicy):
        if cpu_policy == ServerCpuPolicy.one_at_a_time:
            return True
        return False
    
    def is_always_able_to_build_full_path(self):
        return False
    
    def requires_queues_per_class(self):
        return True
    
    def get_load_of_sfis_of_sf(self, sf: int):
        return self.get_arrival_rate_estimate(sf) / self.mySFF.service_rate_per_sf[sf]
    
    def cache_map_server_to_classes(self):
        print(".. cache map server->classes")
        sfi_props: SFI.Props = self.sim.props.sfi
        flow_props: Flow.Props = self.sim.props.flow
        mpp_sched_props: MppScheduler.Props = self.sim.props.mpp_scheduler
        
        assert (mpp_sched_props.map_server_to_classes is None)
        mpp_sched_props.map_server_to_classes = dict()
        if self.sim.DEBUG:
            print("MppScheduler creates map server->classes")
        mpp_sched_props.packet_underway_counter_per_server = {s: 0 for s in self.sim.props.server.all_servers}
        for sfi_id in sfi_props.all_sfi:
            sfi_of_class = sfi_props.all_sfi[sfi_id]
            server = sfi_of_class.server
            if server not in mpp_sched_props.map_server_to_classes:
                affected_classes = []
                # find all sfis that are running on this server
                sf_types = set()
                for sfi in server.SFIs:
                    sf_types.add(sfi.of_type)
                # find all affected traffic classes
                for sfc_class in flow_props.sfc_class_to_sf:
                    sf = flow_props.sfc_class_to_sf[sfc_class][0]
                    if sf in sf_types:
                        affected_classes.append(sfc_class)
                
                mpp_sched_props.map_server_to_classes[server] = affected_classes
        # check if we can reach all queues
        reachable = set()
        for server in mpp_sched_props.map_server_to_classes:
            for queue in mpp_sched_props.map_server_to_classes[server]:
                reachable.add(queue)
        if len(reachable) != flow_props.sfc_next_free_class:
            missing = set(range(flow_props.sfc_next_free_class)).difference(reachable)
            # assert(len(missing)> 0)
            for m in missing:
                missing_sf = flow_props.sfc_class_to_sf[m][0]
                for sfi_id in sfi_props.all_sfi:
                    if sfi_props.all_sfi[sfi_id].of_type == missing_sf:
                        raise NameError(
                            f"I cannot serve all possible queues, but this is strange because there is a sfi")
                if self.assert_on_reject:
                    raise NameError(
                        f"I cannot serve all possible queues, because there is no SFI of type: {missing_sf}, {m}, {flow_props.sfc_class_to_sf[m]}")
            
            if self.assert_on_reject:
                raise NameError(
                    f"I cannot serve all possible queues. Missing: {missing} of {flow_props.sfc_classes} having "
                    f"{flow_props.sfc_class_to_sf}")
        print("... [DONE]")
    
    def is_a_valid_activity(self, activity_id):
        properties = self.get_properties_of_activity(activity_id)
        sf, _ = self.sim.props.flow.sfc_class_to_sf[properties['queue']]
        return sf == self.sim.props.sfi.all_sfi[properties["sfi_id"]].of_type
    
    def get_properties_of_activity(self, activity_id):
        tmp_1 = len(self.sim.props.sff.allSFFs) * self.sim.props.flow.sfc_next_free_class
        sfi_id = int(activity_id / tmp_1)
        remainder = activity_id - tmp_1 * sfi_id
        queue = int(remainder / len(self.sim.props.sff.allSFFs))
        remainder -= queue * len(self.sim.props.sff.allSFFs)
        sff_id = remainder
        return {'sfi_id': sfi_id, 'queue': queue, 'sff_id': sff_id}
    
    def get_activity_for(self, sfi_id: int, queue: int, sff_id_of_packet_class: int) -> int:
        return len(self.sim.props.sff.allSFFs) * self.sim.props.flow.sfc_next_free_class * sfi_id + \
               len(self.sim.props.sff.allSFFs) * queue + \
               sff_id_of_packet_class
    
    def cache_r_matrix(self):
        print(".. cache r matrix")
        sfi_props: SFI.Props = self.sim.props.sfi
        sff_props: SFF.Props = self.sim.props.sff
        flow_props: Flow.Props = self.sim.props.flow
        mpp_sched_props: MppScheduler.Props = self.sim.props.mpp_scheduler
        
        assert (mpp_sched_props.r_matrix is None)
        # how many activities do we have? max(SFI.id) * len(buffers) * len(sff)
        total_activities = self.sim.props.sfi.next_free_id * flow_props.sfc_next_free_class * len(sff_props.allSFFs)
        
        mpp_sched_props.r_matrix = {key: 0 for key in range(total_activities)}
        
        if self.sim.DEBUG:
            print("MppScheduler creates r_matrix")
        
        alpha_normalized_enumerator = math.pow(flow_props.max_deadline, 2)
        
        for sff_id in sff_props.allSFFs:
            sff = sff_props.allSFFs[sff_id]
            for queue in flow_props.sfc_class_to_sf:
                
                # get alpha for this queue, and set r values accordingly
                alpha_denominator = math.pow(flow_props.sfc_class_to_deadline[queue], 2)
                alpha = alpha_normalized_enumerator / alpha_denominator
                
                if not self.consider_alpha_by_using_timeouts:
                    alpha = 1
                
                assert (1 <= alpha)
                
                for sfi_id in sfi_props.all_sfi:
                    sfi = sfi_props.all_sfi[sfi_id]
                    if sfi.server.cpu_policy != ServerCpuPolicy.one_at_a_time:
                        raise NameError("this scheduler does not support other cpu sharing methods than one at a time."
                                        " (we consider R as static)")
                    activity = self.get_activity_for(sfi_id=sfi.id, sff_id_of_packet_class=sff.id, queue=queue)
                    sf, is_last_of_sfc = flow_props.sfc_class_to_sf[queue]
                    if sf == sfi.of_type:
                        # calculate the rate of the activity
                        delay = sfi.get_expected_processing_time()
                        if sfi.sffId != sff.id:
                            delay += sff.get_multi_hop_latency_for(self.sim, source_id=sff.id, dest_id=sfi.sffId)
                        
                        mpp_sched_props.r_matrix[activity] = alpha * 1000000.0 / delay
                        assert (mpp_sched_props.r_matrix[activity] > 0)
                    else:
                        # set to 0, so we dont have to change something
                        pass
        
        if self.sim.DEBUG:
            print(".. with a total of {0} entries".format(len(mpp_sched_props.r_matrix)))
        print("... [DONE]")
    
    def notify_sfi_finished_processing_of_packet(self, sfi: SFI, packet: Packet):
        mpp_sched_props: MppScheduler.Props = self.sim.props.mpp_scheduler
        
        if self.sim.DEBUG:
            print(f"MppScheduler gets notified that sfi {sfi} is finished of packet {packet}, "
                  f"so remove server {sfi.server} from the blocked list")
        
        if 'mpp_locking' in packet.scheduler_flag and packet.scheduler_flag['mpp_locking'] is True:
            mpp_sched_props.packet_underway_counter_per_server[sfi.server] -= 1
            packet.scheduler_flag['mpp_locking'] = False
        
        # check if this was the last event, if so, we do not queue at the sfi,
        # if not, we might queue at the sfi
        if mpp_sched_props.packet_underway_counter_per_server[sfi.server] == 0:
            # since we run always in one at a time policy and we do not queue packets at the SFI, the server HAS TO
            # BE FREE if we are finished with a sfi
            assert (sfi.server.is_free())
        
        if self.block_sfi_while_packet_on_wire:
            assert sfi.free
            mpp_sched_props.blocked_sfi.remove(sfi)
        
        if (mpp_sched_props.allow_up_to_x_packets_underway_per_server -
                mpp_sched_props.packet_underway_counter_per_server[sfi.server] >=
                mpp_sched_props.batch_scheduling):
            # inform all affected schedulers
            sff_props: SFF.Props = self.sim.props.sff
            sff_to_be_informed = sorted(sfi.server.SFF_ids)
            self.sim.random.shuffle(sff_to_be_informed)
            for sff_id in sff_to_be_informed:
                # schedule a scheduling event which terminates now this is required,
                # because we should do scheduling after
                # the callback of this function (control flow of the packet needs to be finished first)
                sff_props.allSFFs[sff_id].scheduler.free_server_count += 1
                self.sim.schedule_event(DoSchedulingEvent(delay=0, scheduler=sff_props.allSFFs[sff_id].scheduler))
    
    def notify_packet_was_dropped(self, packet, caller):
        # we have to get the affected server, so that we release the lock
        sfi = None
        if caller is not None:
            if caller.__class__.__name__ == SFF.__name__:
                # get the sfi from the packet's path
                for path_cls, path_instance in packet.fullPath[packet.pathPosition:]:
                    if path_cls == SFI.__name__:
                        # we found the next sfi, this is the guy we are intrested in
                        sfi = path_instance
                        break
            elif caller.__class__.__name__ == SFI.__name__:
                sfi = caller
            else:
                raise NameError(f"packet was dropped, but caller is neither SFI nor SFF, caller={caller}")
        
        if sfi is not None:
            self.notify_sfi_finished_processing_of_packet(sfi, packet)
        else:
            # nothing found, this is a race condition
            raise NameError("I was not able to find a scheduled SFI in the packet's path, but there should be at least "
                            "one next SFI!")
    
    def apply_scheduling_logic_for_packet(self, packet: Packet):
        ## check if next hop matches available SFIs
        expected_sf = packet.toBeVisited[0]
        
        if self.accessible_sf is None:
            self.accessible_sf = set()
            if self.oracle:
                sfi_props: SFI.Props = self.sim.props.sfi
                for sfi_id in sfi_props.all_sfi:
                    sfi = sfi_props.all_sfi[sfi_id]
                    self.accessible_sf.add(sfi.of_type)
            else:
                for sf in self.mySFF.service_rate_per_sf:
                    if (self.mySFF.service_rate_per_sf[sf] > 0):
                        self.accessible_sf.add(sf)
        
        if (expected_sf not in self.accessible_sf):
            queue = Flow.get_packet_class_of_packet(packet)
            
            popped_packet = self.mySFF.packet_queue_per_class[queue].pop()
            assert (packet == popped_packet)
            packet.reject()
        else:
            self.trigger_scheduling_logic()
    
    def trigger_scheduling_logic(self):
        sfi_props: SFI.Props = self.sim.props.sfi
        sff_props: SFF.Props = self.sim.props.sff
        flow_props: Flow.Props = self.sim.props.flow
        mpp_sched_props: MppScheduler.Props = self.sim.props.mpp_scheduler
        
        # check cached data structure
        if mpp_sched_props.map_server_to_classes is None:
            self.cache_map_server_to_classes()
        
        if mpp_sched_props.r_matrix is None:
            self.cache_r_matrix()
        
        if self.free_server_count == 0:
            return False
        
        if self.sim.DEBUG:
            print("Mpp Scheduler starts scheduling")
        
        if mpp_sched_props.do_sanity_checks:
            packets_queued_before = sum(
                [len(x) for sff in sff_props.allSFFs.values() for x in sff.packet_queue_per_class.values()])
        
        successfully_scheduled = 0
        take_the_next_decision = True
        # we repeat the scheduler, as long as we found a scheduling decision in the previous loop
        # if we are in oracle case, we check all servers, if non oracle, then only the servers of my sff
        servers_to_process_packet: List[Server] = list(
            mpp_sched_props.map_server_to_classes.keys()) if self.oracle else self.mySFF.servers
        
        # now select all sffs, from which we have to check the buffers to take a packet from the queue
        # if oracle, we check all SFF, if non oracle, we check only my own sff
        source_sff_to_check: List[int] = list(sff_props.allSFFs.keys()) if self.oracle else [self.mySFF.id]
        pop_sff: deque[int] = deque()
        
        while take_the_next_decision:
            # now filter this list if we block server on which we already scheduled a packet from somewhere.
            # this is only important when in oracle scheduling case
            servers_to_process_packet = [s for s in servers_to_process_packet
                                         if mpp_sched_props.allow_up_to_x_packets_underway_per_server -
                                         mpp_sched_props.packet_underway_counter_per_server[s] >=
                                         mpp_sched_props.batch_scheduling]
            
            self.free_server_count = len(servers_to_process_packet)
            if self.free_server_count == 0:
                break
            
            # build a list of the best activities I found
            # this map is Server from which I take a packet -> (p_value, activity)
            best_activity_among_all_servers: Dict[Server, Tuple[int, int]] = dict()
            
            # we now measure the real time of scheduling, just for statistics reasons
            self.mark_time_scheduling_starts()
            for server_to_process_sfi in servers_to_process_packet:
                if self.sim.DEBUG:
                    print(".. found a free server {0}".format(server_to_process_sfi))
                
                # activity -> p_value
                p_values: Dict[int, int] = dict()
                
                # this list holds all classes, which could be served by this server (by any of its SFIs)
                affected_classes_of_server: List[int] = mpp_sched_props.map_server_to_classes[server_to_process_sfi]
                
                # now we select each possible SFF from which we could take a packet
                while len(pop_sff) > 0:
                    source_sff_to_check.remove(pop_sff.popleft())
                
                for sff_source_id in source_sff_to_check:
                    sff_source = sff_props.allSFFs[sff_source_id]
                    # and check for this sff each queue,
                    # corresponding to any of the classes that are of interest for this server
                    sff_has_nothing = True
                    for queue in sff_source.packet_queue_per_class:
                        # now we check if there is a packet in this queue
                        
                        while len(sff_source.packet_queue_per_class[queue]) > 0:
                            p = sff_source.packet_queue_per_class[queue][0]
                            if p.flow.qosMaxDelay < self.sim.currentTime - p.time_ingress:
                                # drop this packet
                                packet = sff_source.packet_queue_per_class[queue].popleft()
                                packet.timeQueueScheduling += packet.get_delta_of_time_mark()
                                packet.drop_timed_out(self)
                            else:
                                break
                        
                        if len(sff_source.packet_queue_per_class[queue]) > 0:
                            sff_has_nothing = False
                            if queue in affected_classes_of_server:
                                # yes, so get the SF of this class
                                type_of_class, last_sf_of_sfc = flow_props.sfc_class_to_sf[queue]
                                # select now a corresponding SFI of this server
                                sfi_target: SFI
                                for sfi_target in server_to_process_sfi.SFIs:
                                    # and check if this sfi is of the requested sf type for this class/queue
                                    # and check if (in case of non oracle mode) we are allowed to use this sfi
                                    if sfi_target.of_type == type_of_class and (self.oracle or
                                                                                sfi_target.sffId == sff_source_id):
                                        
                                        # do we block this sfi?
                                        if self.block_sfi_while_packet_on_wire:
                                            if sfi_target in mpp_sched_props.blocked_sfi:
                                                continue
                                        
                                        # no, so get the activity id of this action
                                        activity = self.get_activity_for(sfi_id=sfi_target.id, queue=queue,
                                                                         sff_id_of_packet_class=sff_source.id)
                                        
                                        # get the p value
                                        p_value = self.get_p_value_for(activity=activity)
                                        
                                        if self.sim.DEBUG:
                                            print(f'.... found a possible activity {activity} with p:{p_value}, '
                                                  f'doing {self.get_properties_of_activity(activity)}, with sfi of type '
                                                  f'{sfi_props.all_sfi[self.get_properties_of_activity(activity)["sfi_id"]].of_type} '
                                                  f'for sfc identifier '
                                                  f'{Flow.debug_get_sfc_identifier_and_pos_of_packet_class(self.sim, queue)}')
                                        
                                        # add this as an possible option
                                        p_values[activity] = p_value
                    if sff_has_nothing:
                        pop_sff.append(sff_source_id)
                
                # now we select for this server (from which we could take a packet),
                # the best activity among all possible
                best = self.select_best_activity(p_values)
                if best is not None:
                    # we found any, so lets ensure that we will run the whole scheduling a 2nd time,
                    # just if in this scheduling round we will
                    # select a different server from which we will take a packet
                    take_the_next_decision = True
                    best_activity_among_all_servers[server_to_process_sfi] = (p_values[best], best)
            
            if len(best_activity_among_all_servers) == 0:
                if self.sim.DEBUG:
                    print("... there is no possible activity I could schedule")
                self.reset_timer()
                take_the_next_decision = False
            else:
                # print(best_activity_among_all_servers)
                best_p = None
                best_server = None
                for server_to_process_sfi in best_activity_among_all_servers:
                    if best_p is None or best_activity_among_all_servers[server_to_process_sfi][0] > best_p:
                        best_server = server_to_process_sfi
                        best_p = best_activity_among_all_servers[server_to_process_sfi][0]
                
                best = best_activity_among_all_servers[best_server][1]
                server_to_process_sfi = best_server
                # print(f"select {best_server} with activity {best}")
                
                # schedule a packet from this queue and proceed with the next server
                activity_id = best
                properties = self.get_properties_of_activity(activity_id)
                target_sfi: SFI = sfi_props.all_sfi[properties['sfi_id']]
                target_sff: SFF = sff_props.allSFFs[target_sfi.sffId]
                from_sff: SFF = sff_props.allSFFs[properties['sff_id']]
                from_queue = properties['queue']
                
                assert target_sfi.server == server_to_process_sfi
                
                assert mpp_sched_props.packet_underway_counter_per_server[target_sfi.server] < \
                       mpp_sched_props.allow_up_to_x_packets_underway_per_server
                
                if self.block_sfi_while_packet_on_wire:
                    mpp_sched_props.blocked_sfi.add(target_sfi)
                
                scheduled_path = []
                
                # do we have to go to another SFF?
                if target_sff != from_sff:
                    assert self.oracle
                    path = SFF.get_multi_hop_path_for(self.sim, from_sff.id, target_sff.id)
                    for next_sff in path:
                        scheduled_path.append((SFF.__name__, sff_props.allSFFs[next_sff]))
                
                # push to the SFI
                scheduled_path.append((SFI.__name__, target_sfi))
                
                # push back to the SFF
                scheduled_path.append((SFF.__name__, target_sff))
                
                # how many packet shall we send on this path?
                packet_count = min(mpp_sched_props.batch_scheduling,
                                   mpp_sched_props.allow_up_to_x_packets_underway_per_server -
                                   mpp_sched_props.packet_underway_counter_per_server[target_sfi.server])
                time_delta_of_scheduling = self.get_time_delta_of_scheduling()
                
                self.scheduling_attempts += 1
                
                while packet_count > 0 and len(from_sff.packet_queue_per_class[from_queue]) > 0:
                    # pop a packet from queue and attach path
                    packet = from_sff.packet_queue_per_class[from_queue].popleft()
                    
                    packet.realTimeScheduling += time_delta_of_scheduling
                    packet.timeQueueScheduling += packet.get_delta_of_time_mark()
                    
                    time_left = packet.flow.qosMaxDelay - (self.sim.currentTime - packet.time_ingress)
                    
                    min_time = (self.mySFF.get_multi_hop_latency_for(self.sim,
                                                                       self.mySFF.id, target_sff.id)
                                + self.mySFF.get_multi_hop_latency_for(self.sim,
                                                                       target_sff.id, packet.flow.desiredEgressSSFid))

                    if time_left < min_time:
                        if self.sim.DEBUG:
                            print(". drop packet because of timeout inside scheduling queue")
                        
                        packet.drop_timed_out(self)
                        continue
                    
                    packet_count -= 1
                    
                    packet.set_callback_when_dropped(lambda p, caller: self.notify_packet_was_dropped(p, caller))
                    
                    mpp_sched_props.packet_underway_counter_per_server[target_sfi.server] += 1
                    packet.scheduler_flag['mpp_locking'] = True
                    
                    
                    expected_sf = packet.toBeVisited.pop(0)
                    if expected_sf != target_sfi.of_type:
                        raise NameError("the scheduler messed up something!")
                    
                    packet.fullPath += scheduled_path
                    
                    if packet.id == self.sim.PACKET_ID_TO_DEBUG:
                        print(
                            f"** packet was scheduled from sff {from_sff} queue {from_queue} "
                            f"to sff {target_sff} on sfi {target_sfi} running on server {target_sfi.server} ")
                    
                    if self.sim.DEBUG:
                        print(f"... select activity {best} with p {best_p} from sff {from_sff} "
                              f"queue {from_queue} to sff {target_sff} on sfi {target_sfi} ")
                        Packet.debug_packet(packet)
                    
                    successfully_scheduled += 1
                    from_sff.handle_packet_from_scheduler(packet)
        
        # do sanity check?
        if mpp_sched_props.do_sanity_checks:
            if successfully_scheduled == 0 and not self.oracle:
                
                # check if non of my sfis is free, and if so, then there should be no packet in any of the queues
                for sf in self.mySFF.SFIsPerType:
                    waiting_queues_for_sf = []
                    for queue in self.mySFF.packet_queue_per_class:
                        if 0 < len(self.mySFF.packet_queue_per_class[queue]):
                            # there are packets in this queue, check the sf type of this queue
                            sfc, pos = Flow.debug_get_sfc_identifier_and_pos_of_packet_class(self.sim, queue)
                            flow_sf, eos = Flow.get_sf_and_eoc_of(self.sim, sfc, pos)
                            if flow_sf == sf:
                                waiting_queues_for_sf.append(queue)
                    
                    # if there is no queue waiting, we don't have to check the sfis
                    if 0 == len(waiting_queues_for_sf):
                        continue
                    
                    for sfi in self.mySFF.SFIsPerType[sf]:
                        if sfi.free and sfi.server.is_free():
                            raise NameError(f'there is a free sfi {sfi} of a sff {self.mySFF}, '
                                            f'which could serve a packet from a queue ('
                                            f'{[(queue, Flow.debug_get_sfc_identifier_and_pos_of_packet_class(self.sim, queue)) for queue in waiting_queues_for_sf]}).')
        
        return successfully_scheduled > 0
    
    def get_p_value_for(self, activity: int):
        sfi_props: SFI.Props = self.sim.props.sfi
        sff_props: SFF.Props = self.sim.props.sff
        flow_props: Flow.Props = self.sim.props.flow
        mpp_sched_props: MppScheduler.Props = self.sim.props.mpp_scheduler
        
        properties = self.get_properties_of_activity(activity)
        source_sff = sff_props.allSFFs[properties["sff_id"]]
        target_sfi = sfi_props.all_sfi[properties["sfi_id"]]
        target_sff = sff_props.allSFFs[target_sfi.sffId]
        queue = properties["queue"]
        
        expected_sf, end_of_sfc = flow_props.sfc_class_to_sf[queue]
        
        assert expected_sf == target_sfi.of_type
        assert queue in source_sff.packet_queue_per_class
        
        # here, we subtract the length of the "on the way packets"
        # rate of this activity times the queue length at source sff
        
        subtract_sfi_queue = 0
        
        p_value = mpp_sched_props.r_matrix[activity] * (
                len(source_sff.packet_queue_per_class[queue]) -
                mpp_sched_props.packet_underway_counter_per_server[target_sfi.server] -
                subtract_sfi_queue)
        
        # backpressure?
        if not end_of_sfc:
            next_queue = queue + 1
            if next_queue in target_sff.packet_queue_per_class:
                p_value -= mpp_sched_props.r_matrix[activity] * len(target_sff.packet_queue_per_class[next_queue])
        
        return p_value
    
    def select_best_activity(self, p_values: Dict[int, float]):
        sff_props: SFF.Props = self.sim.props.sff
        
        # find the best activity
        best: Tuple[float, int, int] = (None, None, None)  # p_val, activity, time_when_packet_entered_sff
        for activity, p_value in p_values.items():
            # we select the max value, or
            #   if values are equally, we take the activity which has the older packet in the queue,
            #   but packet time with respect to time when packet entered sff for queueing
            #       if also the time is equal, then we select the activity with the largest id
            properties = self.get_properties_of_activity(activity)
            time_of_packet = sff_props.allSFFs[properties['sff_id']].packet_queue_per_class[properties['queue']][
                0].timeMarker
            if best[0] is None or best[0] < p_value or (
                    best[0] == p_value and (best[2] > time_of_packet or (
                    best[2] == time_of_packet and best[1] < activity))):
                # get time of packet
                best = (p_value, activity, time_of_packet)
        return best[1]
    
    def debug_print_r_matrix(self):
        sfi_props: SFI.Props = self.sim.props.sfi
        sff_props: SFF.Props = self.sim.props.sff
        flow_props: Flow.Props = self.sim.props.flow
        mpp_sched_props: MppScheduler.Props = self.sim.props.mpp_scheduler
        
        for random_sff in sff_props.allSFFs.values():
            if isinstance(random_sff.scheduler, MppScheduler):
                random_sff.scheduler.cache_r_matrix()
                break
        
        for queue in range(len(flow_props.sfc_class_to_sf)):
            sf = flow_props.sfc_class_to_sf[queue][0]
            print(f" packet class {queue},  which refers to SF {sf}")
            
            options = dict()
            
            for dest_sff in sff_props.allSFFs.values():
                for sfi in sfi_props.all_sfi.values():
                    if sfi.of_type == sf and sfi.sffId == dest_sff.id:
                        # I found a sfi at the dest sff which could serve this queue
                        # now loop through all source SFFs from which I could take a packet
                        for source_sff in sff_props.allSFFs.values():
                            activity = source_sff.scheduler.get_activity_for(sfi_id=sfi.id,
                                                                             sff_id_of_packet_class=source_sff.id,
                                                                             queue=queue)
                            assert source_sff.scheduler.is_a_valid_activity(activity)
                            if source_sff not in options:
                                options[source_sff] = []
                            options[source_sff].append(activity)
            
            for source_sff in options:
                print(f"\t taking a packet from SFF {source_sff.id}")
                print(f"\t\t {[round(mpp_sched_props.r_matrix[x], 1) for x in options[source_sff]]}")
